    %include "Include\ProtoS.inc"

; ============================================================================================
; At this juncture, none of the registers has any meaningful data except

;       SS:BP + 0 = Maximum number of entries FF if there was and error and
;                   equal to COUNT (BP+1) if overflow
;             + 1 = Number of entries in E820 map (32 bytes, only 20 usually used though)
;             + 2 = Long pointer to E820, probably (0x500). see MAP_BASE in Stork.inc

;             + 6 =  2 (Green)  A20 line was already on.
;                   14 (Yellow) Fast A20 gate succeeded.
;                    4 (Red)    A20 activate failed.
;             
;       SS:BP - 04 = Far pointer back into BIOS code (RETF essentially)

;       Contents of stack as written in pre-amble of Boot.asm
;                    AX  CX  DX  BX  SP  BP  SI  DI  CS  DS  ES  SS  FS  GS
;            SS:BP - 06  08  0A  0C  0E  10  12  14  16  18  1A  1C  1E  20 -< SS:SP is here.
; --------------------------------------------------------------------------------------------

  Begin:
	mov	ax, cs			; CS has been established by far jump from boot sect.
	mov	ds, ax
	mov	es, ax
	
    ; In the event BIOS hasn't cleared screen. BOCH's is one such example, but my testbed
    ; Lenovo ThinkCenter does.

	mov	ax, 3
	int	VIDEO

    ; Display OS name, centered on 2nd line.

	push	word GREEN
	push	word 0x21e		; If string is changed, this need to be changed
	push	Prompt
	call	ShowS

    ; Display secondary prompt that is pointed to by AX now

	push   word RED
	push   word 0x50a		; Position @ row 5 column 10
	push   ax			; Points to string after "Prompt"
	call   ShowS
	mov    si, ax			; Pointer to register ident pairs.

    ; Draw a single white line under last text

	mov	bx, WHITE		; Attribute for line and so BH = 0 (page 1)
	mov	ax, SCP * 256
	mov	dx, 0x509		; Needs to be zero indexed.
	int	VIDEO

	mov	ax, WCA * 256 + 0xcd	; Solid underline
	mov	cx, 27			; Should really do this with an equate
	int	VIDEO			; Draw line, don't bother moving cursor

; ============================================================================================
; Cycle through and display each of the 16 registers that were saved in bootloader preamble
; --------------------------------------------------------------------------------------------

    ; Setup for display loop

	mov	di, Fmt			; Pointer to 16 register mnemonics
	mov	dx, 0x80a		; Begin @ row 8, column 10
	lea	bx, [bp - 2]		; BX points to beginning of list.

  .NextSet:
	lodsw				; Read next nmemonic
	cmp	ax, -1			; -1 marks end of list
	jz	.Done

	mov	[di], ax		; Move reg ident into first characters of output
	push	ax			; Need this for tabulating display @ Line 113

    ; Show first 5 characters of display in GREEN

	push	word GREEN		; Arg3 = Attribute and page (0 - 7) @246@
	push	dx			; Arg2 = Beginning coordinate
	push	di			; Arg1 = Pointer to string to ASCIIZ string
	call	ShowS
	sub	sp, 6			; Restore arguments to ShowS

    ; Setup for ShowS

	push	word LRED		; @251@
	push	-1			; Carry on from last position
	push	ax			; Set for next string

    ; Bump pointer to end of ASCIIZ string and convert value at [BX]

	push	word 0x120		; Padd output with spaces / 16 bit conversion
	add	ax, 3
	push	ax			; ARG 2 Pointer to end of string
	mov	ax, [ss:bx]
	push	eax			; ARG 1 16/32 bit value
	or	ax, ax			; Set ZR flag
	call	I2H			; Convert to hex (NULLS will be padded space)
	jnz	.J0			; Was last conversion zero
	
	mov	byte [bp-42], DGREY	; Zeros are displayed in dark grey @26B@
			
   .J0:	call	ShowS			; Display 16 bit value

    ; If it wasn't for the fact this code is only executed once, I would have put a little
    ; more thought into my design as this is essentially only here to preserve DX
    
	pop	ax			; Waste as contents of DI already equals this
	pop	dx			; Restore last position
	inc	dh			; Bump to next line
	pop	ax			; Waste AX so stack is in proper position again

    ; Bump one line when "IP" is encountered

	pop	ax			; Restore 2 byte register identifier
	cmp	ax, "IP"
	jnz	.00

	inc	dh			; Bump line count just before "AX"
	jmp	.Cont

    ; Move to a column to the right, so right side of output is aligned with double line
    ; above it.

   .00: cmp	ax, "DI"
	jnz	.Cont
	mov	dx, 0x0b1c		; Calculated manually.

    ; Point to next value in stack

 .Cont: dec	bx
	dec	bx			; Work backward through list of values
	jmp	.NextSet		; Carry on till [BX] points to -1

    ; Draws a single line under everything now

 .Done: mov	bx, WHITE		; Attribute for line @28B@
	mov	ax, SCP * 256
	mov	dx, 0x1309		; Needs to be zero indexed.
	int	VIDEO

	mov	ax, WCA * 256 + 0xcd	; Solid underline
	mov	cx, 27			; Should really do this with an equate
	int	VIDEO			; Draw line, don't bother moving cursor
	mov	sp, bp			; Values on stack no longer needed.

; ============================================================================================
; Display "A20 Line" in color determined in bootloader code.

	;  GREEN = A20 was already ON
	; YELLOW = A20 was turned ON using fast method through port 92H
	;    RED = Failed to turn gate on, alternate method must be implemented in Boot.
; --------------------------------------------------------------------------------------------
    ; Display "A20 Line" in color specified at A20_LINE

	xor	ax, ax			; Sets high nibble to nulls
	mov	al, [A20_LINE]		; Color GREEN/YELLOW/RED
	push	ax
	add	dx, 0x301		; Because current DX is zero indexed
	push	dx
	push	word A20		; Pointer to ASCIIZ string
	call	ShowS

    ; Display "TOM", top of memory. This is the value extracted from 40:13 in BDA and
    ; converted 32 bit address
    
	push	word LGREY		; @2B1@
	sub	sp, 2
	pop	dx			; Previous DX from line 161
	add	dl, 17
	push	dx
	push	ax
	call	ShowS	

    ; A stack frame of 65,536 bytes was initialized in bootloader, so this is why we
    ; can use SS to establish where TOM 40:13 in BDA was
    	
	dec	ax
	dec	ax			; Point to last character of string @ A20
	push	word 0x200		; 32 bit conversion, no padding
	push	ax			; Pointer to last character of string
	mov	ax, ss
	add	ah, 16			; Because we know stack is 64K
	shl	eax, 4			; Make 32 bits
	push	eax
	call	I2H
	inc	ax			; To beginning of ASCIIZ string
	
	push	word CYAN		; @2D4"
	push	word -1
	push	ax
	call	ShowS
	
; ============================================================================================
; Display E820 map. This implementation assumes there is 0<20 entries so it would have to be
; reworked in some way, especially if there were more than 25
; --------------------------------------------------------------------------------------------

	lfs	bx,[E820_PNTR]		; Pointer to E820 map in FS:BX
	mov	di, 0xd30		; Begin display @ Row 13, Column 48
	
    ; Determine where the first line should start by Line 13 - (count/2) + 1
    
	mov	cl, [COUNT]		; Get number of entries in map
	mov	ax, cx
	shl	ax, 7
	bt	ax, 8
	jnc	$ + 4			; Bump down one line if AX was odd
	dec	ah
	sub	di, ax
	
    ; @2f6@ is where you could change pointing characters, but you are limited to 4
    
	mov	eax, 0x2010fe
	mov	[si], eax
	mov	[si-1], byte 0
	jmp	ShowMap

    ; These are here just because.
    
    %include	'Show16.asm'
    align	16			; No particular reason for alignment
    %include	'I2H_16.asm'			

; --------------------------------------
; This routine takes displays 64 bit value in map and converts to 48 bit display.
; --------------------------------------	

  Build_Str:
	push	di
	push	ds			; Preserve
	push	fs
	pop	ds
	xchg	si, di
	mov	si, bx
	add	si, 12			; DS:SI bits 63 - 48 of map value
	std				; Move backward through buffer
	
    ; Setup stack to convert upper half
    
   .L0:	mov 	cx, 0x120		; Upper half is 16 bit, padded with spaces
	push	cx
	lodsd
	push	eax			; Saved temporarily for call to I2H later
	or	eax, eax
	jz	.J0
	or	cl, 16			; Convert fill byte from spaces to zeros
	
    ; Setup stack to convert lower half
    
   .J0:	bts	cx, 9			; Set bit to do 32 bit conversion now
	lodsd				; Get lower half of value
	push	cx			; ARG3: Conversion type and fill byte
	push	di			; ARG2: Pointer to last position of output
	push	eax			; ARG1: 32 bit value (default) to be converted
	call	I2H
	
    ; Convert upper half now
    					; ARG3: already set @ line 243
	pop	ecx			; Recover 32 bit value to be converted
	push	ax			; ARG2: Next position in output buffer
	push	ecx			; ARG1: 32 bit value
	call	I2H
	
    ; At minimum there is always at least one zero displayed for null values, but in this
    ; case we want to eliminate it, if that's the case.
    
	mov	di, ax
	cmp	word [es:di+3], ' 0'
	jnz	.J1
	mov	byte [es:di+4], ' '
	
   .J1:	cmp 	bx, si
	jg	.Done			; This works because we know map always starts @ 0x500
	dec	di
	mov	word [es:di], 0xba20	; Deliniate base address and size with vertical bars
	dec	di
	jmp	.L0

    ; Cleanup and return pointer to ASCIIZ string in AX	
 .Done:	pop	ds
 	pop	di
 	inc	ax
 	cld
 	ret
 	
; Cycle through each of the map entries, displaying Base Address / vertical bars / Size

  ShowMap:
	push	cx	
	mov	al, [fs:bx+16]		; Get entry type
	
    ; 1 = GREEN, 2 = DGREY, 3 = YELLOW and everything else is RED.
    
	mov	cl, LGREEN
	cmp	al, 1
	jz	.J0
	mov	cl, DGREY
	cmp	al, 2
	jz	.J0
	mov	cl, YELLOW
	cmp	al, 3
	jz	.J0
	mov	cl, RED
	
    ; Pointer to prompt (SI) was established in line 215	
   .J0:	push	cx   	
	push	di
	push	si
	call	ShowS
	push	si
	
	dec	si
	dec	si
	
    ; Build ASCIIZ string Base address & size. 48 bits + will be truncated.
    
	call	Build_Str
	push	word 7
	push	word -1
	push	ax
	call	ShowS
	
    ; Adjust pointers to next map entry
    
	pop	si
	add	bx, 32			; Point to next map entry
	add	di, 256  	
	pop	cx
	loop	ShowMap  	
  	
	jmp	Finish			; Bounce over local subs and strings for BOCH's

	align	16
;  *--*  *--*  *--*  *--*  *--*  *--*  *--*  *--*  *--*  *--*  *--*  *--*  *--*  *--*  *--*

  Prompt:	db	'Proto-System V1.00.0', 0
		db	'Registers as passed by BIOS', 0	; 27 chars
		db	'CSIPAXCXDXBXSPBPSIDICSDSESSSFSGS', 255, 255
     Fmt:	db	'?? - ', 0, '0000', 0
     A20:	db	'A20 Line', 0, 'TOM: ', 0

	times	(1018-$+$$) db 144	; Padd to beginning of sector 4

    ; This is here temporarily so when testing on real hardware any keypress will re-boot
    ; computer.
    	
  Finish:
  	xor	ax, ax
  	int	KEYBOARD		; Wait for some kind of input from operator
  	int	ReBoot